/* socks-gw.c */

#include "socks-gw.h"

static	char	RcsId[] = "$Header: /usr/home/ygk/fwtk/socks-gw/socks-gw.c,v 1.1 2000/05/22 15:07:10 root Exp root $";

static	int	acceptrule();
static	int	net_write();

static	char	rhost[512];
static	char	raddr[512];
static  time_t	start_time;
static  time_t	end_time;

int	timeout = 2 * 60;
char	service[512] = "socks-gw";
char	eaddr[512];


static	struct sockaddr_in	nulldst = {0,0,0};
static	struct socks_request	sreq;
static	struct socks_reply	sreply;

main(ac,av)
int	ac;
char	*av[];
{
	Cfg	*cfp;
	Cfg	*cf;

#ifndef	LOG_DAEMON
	openlog(service,LOG_PID);
#else
	openlog(service,LOG_PID|LOG_NDELAY,LFAC);
#endif

	if (ac > 2 && !strcmp(av[1],"-daemon")) {
		int port = str_to_port(av[2]);

		ac -= 2;
		av += 2;
		syslog(LLEV,"Starting daemon mode on port %d",port);
		do_daemon(port);
	}
	
	if (ac > 2 && !strcmp(av[1],"-as")) {
		strncpy(service,cf->argv[2],sizeof(service));
		ac -= 2;
		av += 2;
		syslog(LLEV,"Starting %.512s",service);
	}

	time(&start_time);

	if((cfp = cfg_read(service)) == (Cfg *)-1) {
		syslog(LLEV,"fwtkcfgerr: %.512s exiting - cannot read configuration",
		service);
		exit(1);
	}

	if((cf = cfg_get("timeout",cfp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: timeout must have one parameter, line %d",
			cf->ln);
			exit(1);
		}
		timeout = atoi(cf->argv[0]);
	}

	if((cf = cfg_get("external",cfp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: external must have one parameter, line %d",
			cf->ln);
			exit(1);
		}
		strncpy(eaddr,cf->argv[0],sizeof(eaddr));
	}

	bzero(&sreq, sizeof(sreq));
	bzero(&sreply, sizeof(sreply));

	if(peername(0,rhost,raddr,sizeof(rhost))) {
		syslog(LLEV,"fwtksyserr: %.512s exiting - cannot get peername",
		service);
		exit(1);
	}

	if(socks_request(0,&sreq)) {
		syslog(LLEV,"fwtksyserr: %.512s exiting - bad SOCKS request",
		service);
		exit(1);
	}

	/* get list of permitted/denied hosts */
	for(cf = cfg_get("hosts",cfp); cf != (Cfg *)0; cf = cfg_get("hosts",(Cfg *)0)) {
		int	x;

		for(x = 0; x < cf->argc; x++) {

			/*
			if we hit a value with a leading '-' we are
			past processing the host names
			*/
			if(!strcmp(cf->argv[x],"-bind"))
				break;
			if(!strcmp(cf->argv[x],"-connect"))
				break;
			if(!strcmp(cf->argv[x],"-dest"))
				break;
			if(!strcmp(cf->argv[x],"-port"))
				break;

			if(hostmatch(cf->argv[x],raddr)) {
				acceptrule(service,cf,cf->argc,cf->argv);
				break;
			}
		}
	}

	exit_deny();
	exit(1); /* not reached */
}

int
exit_deny()
{
	if (sreq.host[0] != '\0' ) {
		syslog(LLEV,"deny host=%.512s/%.20s dest=%.512s port=%.10s",
			rhost, raddr, sreq.host, sreq.port);
	} else {
		syslog(LLEV,"deny host=%.512s/%.20s",rhost,raddr);
	}
	exit(1);
}

static	int
acceptrule(srv,c,ac,av)
char	*srv;
Cfg	*c;
int	ac;
char	**av;
{
	int		len;
        fd_set          rfds;
        fd_set          xfds;
        fd_set          iced;
        int             ib = 0;
        int             ob = 0;
        int             x;
        int             serfd;
	int		privport = 0;
	struct timeval	timo;
	int		found = 1;
static	char		buf[1024 * 4];

	for (x = 0; x < ac; x++) {
		if (!strcmp(av[x],"-dest")) {
			found = found & 2;
			for(x++; x < ac; x++) {
				if (!strcmp(av[x],"{"))
					continue;
				if (!strcmp(av[x],"}"))
					break;
				if(av[x][0] == '-') {
					x--;
					break;
				}
				if ((found & 2) == 0) {
					if (av[x][0] == '!') {
						if (!hostmatch(&av[x][1], inet_ntoa(sreq.dst.sin_addr)))
							found |= 2;
					} else if(hostmatch(av[x], inet_ntoa(sreq.dst.sin_addr)))
						found |= 2;
				}
			}
		} else if (!strcmp(av[x], "-port")) {
			x++;
			if (x >= ac) {
				syslog(LLEV,"fwtkcfgerr: -port takes an argument, %d",c->ln);
				exit(1);
			}
			if (!namatch(av[x], sreq.port))
				return 0;
		} else if (!strcmp(av[x], "-connect")) {
			x++;
			if (sreq.cmd != SOCKS_CONNECT)
				return 0;
		} else if (!strcmp(av[x], "-bind")) {
			x++;
			if (sreq.cmd != SOCKS_BIND)
				return 0;
		}
	}
	if (!found)
		return 0;

	/* src and dst match so now apply this rule */
	if(c->flags & PERM_DENY) {
		exit_deny();
		exit(1);
	}


	if (sreq.ver == SOCKS4_REQUEST_VERSION) {
		sreply.ver = SOCKS4_REPLY_VERSION;
	} else if (sreq.ver == SOCKS5_REQUEST_VERSION) {
		sreply.ver = SOCKS5_REPLY_VERSION;
	} else {
		exit_deny();
		exit(1);
	}

	switch(sreq.cmd)
	{
	case SOCKS_CONNECT:
		sreply.bnd = sreq.dst;
		sreply.atyp = sreq.atyp;
        	if((serfd = socks_connect(&sreply)) < 0) {
               		syslog(LLEV,"cannot connect to server %.512s/%d: %m",
				sreq.host, sreq.host);
			socks_fail(0, &sreply);
               		exit_deny();
			exit(1);
        	}
		if (socks_success(0, &sreply))
			exit(1);
        	syslog(LLEV,"connect host=%.512s/%.20s destination=%.20s port=%.20s",
			rhost, raddr, sreq.host, sreq.port);
		break;
	case SOCKS_BIND:
		sreply.bnd = sreq.dst;
		sreply.atyp = sreq.atyp;
        	if ((serfd = socks_listen(&sreply)) < 0) {
               		syslog(LLEV,"socks_listen failed, %m");
			socks_fail(0, &sreply);
               		exit_deny();
			exit(1);
		}
		if (socks_success(0, &sreply)) {
			exit(1);
		}
		if ((serfd = socks_accept(serfd, &sreply)) < 0) {
               		syslog(LLEV,"socks_accept failed, %m");
			socks_fail(0, &sreply);
               		exit_deny();
			exit(1);
		}
		if (!hostmatch(sreq.host, sreply.host)) {
			socks_fail(0, &sreply);
               		exit_deny();
			exit(1);
		}
		if (socks_success(0, &sreply)) {
			exit(1);
		}
        	syslog(LLEV,"bind from=%.20s to=%.512s/%.20s port=%d",
			sreply.host, rhost, raddr, sreq.port);
		break;
	default:
        	syslog(LLEV,"unsupported command host=%.512s/%.20s destination=%.20s port=%.20s",
			rhost, raddr, sreq.host, sreq.port);
		socks_fail(0, &sreply);
		exit_deny();
		exit(1);
	}

	FD_ZERO(&iced);
        FD_SET(0,&iced);
        FD_SET(serfd,&iced);

        while(FD_ISSET(0, &iced) || FD_ISSET(serfd, &iced)) {
                (void)bcopy(&iced,&rfds,sizeof(fd_set));
                (void)bcopy(&iced,&xfds,sizeof(fd_set));

                timo.tv_sec = timeout;
                timo.tv_usec = 0;

                if(select(serfd + 1,&rfds,(fd_set *)0,&xfds,&timo) <= 0)
			break;

                if(FD_ISSET(0,&rfds)) {
                        if((x = recv(0,buf,sizeof(buf),0)) <= 0) {
                                FD_CLR(0, &iced);
                                shutdown(serfd, 1);
                        } else {
                                if(net_write(serfd,buf,x,0) != x)
                                        break;
                        }
                        ob += x;
                }
                if(FD_ISSET(serfd,&rfds)) {
                        if((x = recv(serfd,buf,sizeof(buf),0)) <= 0) {
                                FD_CLR(serfd, &iced);
                                shutdown(0, 1);
                        } else {
                                if(net_write(0,buf,x,0) != x)
                                        break;
                        }
                        ib += x;
                }
                if(FD_ISSET(0, &xfds)) {
                        if ((x = recv(0, buf, sizeof(buf), MSG_OOB)) < 0) {
                                /* ignore OOB read errors */
                        } else {
                                if (net_write(serfd, buf, x, MSG_OOB) != x)
                                        break;
                        }
                        ob += x;
                }
                if(FD_ISSET(serfd, &xfds)) {
                        if ((x = recv(serfd, buf, sizeof(buf), MSG_OOB)) < 0) {
                                /* ignore OOB read errors */

                        } else {
                                if (net_write(0, buf, x, MSG_OOB) != x)
                                        break;
                        }
                        ib += x;
                }
        }
        time(&end_time);
        syslog(LLEV,"disconnect host=%.512s/%.20s destination=%.20s port=%.20s in=%d out=%d duration=%d",
		rhost, raddr, sreq.host, sreq.port, ib, ob, end_time - start_time);
        exit(0);
}

static int
net_write(s, buf, len, flags)
int s;
char *buf;
int len;
int flags;
{
        int nbytes = 0;
        int j;

        while (nbytes < len) {
                j = send(s, buf, len-nbytes, flags);
                if (j <= 0)
                        return (nbytes ? nbytes : j);
                buf += j;
                nbytes += j;
        }
        return nbytes;
}

int socks_request(s, req)
int	s;
struct socks_request	*req;
{
	if(socks_read(s, &(req->ver), sizeof(req->ver))) {
		syslog(LLEV,"fwtksyserr: socks_read() error");
		exit(1);
	}
	if(req->ver == SOCKS4_REQUEST_VERSION)
		return(socks4_request(s, req));
	if(req->ver == SOCKS5_REQUEST_VERSION)
		return(socks5_request(s, req));
	return(1);
}

int socks_fail(s, reply)
int	s;
struct socks_reply	*reply;
{
	if(reply->ver == SOCKS4_REPLY_VERSION) {
		reply->rep = SOCKS4_FAILED;
		return(socks4_reply(s, reply));
	}
	if(reply->ver == SOCKS5_REPLY_VERSION) {
		reply->rep = SOCKS5_FAILURE;
		return(socks5_reply(s, reply));
	}
	return(0);
}

int socks_success(s, reply)
int	s;
struct socks_reply	*reply;
{
	if(reply->ver == SOCKS4_REPLY_VERSION) {
		reply->rep = SOCKS4_GRANTED;
		return(socks4_reply(s, reply));
	}
	if(reply->ver == SOCKS5_REPLY_VERSION) {
		reply->rep = SOCKS5_SUCCEEDED;
		return(socks5_reply(s, reply));
	}
	return(0);
}
